# -*- coding: utf-8 -*-
#
# Copyright 2017 Ricequant, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum


class CachePolicy(Enum):
    NORMAL = 1,
    MINIMUM = 2,
    EAGER = 3,


_policy = CachePolicy.NORMAL


def set_cache_policy(policy):
    global _policy
    _policy = policy


MULTIPLIER = {
    CachePolicy.NORMAL: 1,
    CachePolicy.MINIMUM: 1 / 16,
    CachePolicy.EAGER: 2,
}


def get_entry_count(count):
    if count is None:
        if _policy == CachePolicy.MINIMUM:
            return 512
        return None

    multiplier = MULTIPLIER[_policy]
    return int(count * multiplier)



# -*- coding: utf-8 -*-
#
# Copyright 2017 Ricequant, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six
import pandas as pd
from collections import OrderedDict, defaultdict

from ..events import Events
from ..const import ACCOUNT_TYPE
from ..utils.repr import properties
from ..utils import safe_round
from ..execution_context import ExecutionContext


class ResultAggregator(object):
    def __init__(self, env, risk_cal):
        env.event_bus.add_listener(Events.POST_SETTLEMENT, self._collect_daily)
        env.event_bus.add_listener(Events.TRADE, self._collect_trade)
        env.event_bus.add_listener(Events.ORDER_CREATION_PASS, self._collect_order)
        self._env = env
        self._orders = defaultdict(list)
        self._trades = defaultdict(list)

        self._portfolios = OrderedDict()
        self._portfolios_dict = defaultdict(OrderedDict)

        self._risk_free_rate = env.data_proxy.get_risk_free_rate(env.config.base.start_date,
                                                                 env.config.base.end_date)
        self._risk_cal = risk_cal

    def _collect_trade(self, account, trade):
        self._trades[trade.trading_datetime.date()].append(trade)

    def _collect_order(self, account, order):
        self._orders[order.trading_datetime.date()].append(order)

    def _collect_daily(self):
        try:
            benchmark_daily_returns = self._env.accounts[ACCOUNT_TYPE.BENCHMARK].portfolio.daily_returns
        except KeyError:
            benchmark_daily_returns = 0

        date = self._env.calendar_dt.date()
        portfolio = self._env.account.get_portfolio(date)
        self._portfolios[date] = portfolio

        self._risk_cal.calculate(date, portfolio.daily_returns, benchmark_daily_returns, self._risk_free_rate)

        # FIXME 千万不要删！会有严重后果！ET：我的锅
        tmp = self._risk_cal.risk

        # 单独分账号保存自己的
        for account_type, account in six.iteritems(self._env.accounts):
            self._portfolios_dict[account_type][date] = account.get_portfolio(date)

    def get_result_dict(self, strategy_name):
        """
        {
            summary[] - dict
            trades[] - dataframe
            positions[] - dataframe
            portfolios[] - dataframe
        }
        <— result_dict
        """
        from enum import Enum

        data_proxy = self._env.data_proxy

        def safe_convert(value, ndigits=3):
            value = safe_round(value, ndigits)
            if isinstance(value, Enum):
                value = str(value).split(".")[1]

            return value

        def symbol(order_book_id):
            return data_proxy.instruments(order_book_id).symbol

        def filter_keys(keys, blacklist=[]):
            return not key.startswith("_") and not key.endswith("_") and key not in blacklist

        # summary
        date, latest_portfolio = list((key, value) for key, value in six.iteritems(self._portfolios))[-1]
        risk = self._risk_cal.daily_risks[date]
        risk_keys = [
            "volatility", "max_drawdown",
            "alpha", "beta", "sharpe",
            "information_ratio", "downside_risk",
            "tracking_error", "sortino",
        ]

        summary = {
            "strategy_name": strategy_name,
        }
        for key, value in six.iteritems(self._env.config.base.__dict__):
            if key not in ["trading_calendar", "account_list", "timezone", "persist_mode",
                           "resume_mode", "data_bundle_path", "handle_split", "persist", "cal_risk_grid"]:
                summary[key] = safe_convert(value, 2)
        for key in risk_keys:
            summary[key] = safe_convert(getattr(risk, key), 3)
        for key, value in six.iteritems(properties(latest_portfolio)):
            if key not in ["positions", "daily_returns", "daily_pnl"]:
                summary[key] = safe_convert(value, 3)
        # summary["start_date"] = summary["start_date"].strftime("%Y-%m-%d")
        # summary["end_date"] = summary["end_date"].strftime("%Y-%m-%d")

        # add benchmark
        if ACCOUNT_TYPE.BENCHMARK in self._portfolios_dict:
            _, latest_benchmark_portfolio = list((key, value) for key, value in six.iteritems(self._portfolios_dict[ACCOUNT_TYPE.BENCHMARK]))[-1]
            for key in ["total_returns", "annualized_returns"]:
                summary["benchmark_{}".format(key)] = getattr(latest_benchmark_portfolio, key)

        trades_list = []
        total_portfolios_list = []
        positions_list_dict = defaultdict(list)
        portfolios_list_dict = defaultdict(list)
        portfolio_blacklist = ["positions", "start_date", "starting_cash"]

        # total portfolios
        for date, portfolio in six.iteritems(self._portfolios):
            # portfolio
            # dict_data = {"date": date.strftime("%Y-%m-%d")}
            dict_data = {"date": date}
            for key, value in six.iteritems(properties(portfolio)):
                if filter_keys(key, portfolio_blacklist):
                    dict_data[key] = safe_convert(value, 3)
            total_portfolios_list.append(dict_data)

        # stock/future portfolio
        for account_type, portfolio_order_dict in six.iteritems(self._portfolios_dict):
            positions_list = positions_list_dict[account_type]
            portfolios_list = portfolios_list_dict[account_type]

            for date, portfolio in six.iteritems(portfolio_order_dict):
                # portfolio
                # dict_data = {"date": date.strftime("%Y-%m-%d")}
                dict_data = {"date": date}
                for key, value in six.iteritems(portfolio.__dict__):
                    if filter_keys(key, portfolio_blacklist):
                        dict_data[key] = safe_convert(value, 3)
                portfolios_list.append(dict_data)

                # positions
                for order_book_id, position in six.iteritems(portfolio.positions):
                    dict_data = {
                        # "date": date.strftime("%Y-%m-%d"),
                        "date": date,
                    }
                    for key, value in six.iteritems(position.__dict__):
                        if filter_keys(key):
                            dict_data[key] = safe_convert(value, 3)
                    dict_data["order_book_id"] = order_book_id
                    dict_data["symbol"] = symbol(order_book_id)
                    positions_list.append(dict_data)

        # trades
        for date, _ in six.iteritems(self._portfolios):
            for trade in self._trades[date]:
                dict_data = {
                    "side": safe_convert(trade.order.side),
                }
                for key, value in six.iteritems(properties(trade)):
                    if filter_keys(key, ["order"]):
                        dict_data[key] = safe_convert(value, 3)
                dict_data["order_book_id"] = trade.order.order_book_id
                dict_data["symbol"] = symbol(trade.order.order_book_id)
                dict_data["datetime"] = dict_data["datetime"].strftime("%Y-%m-%d %H:%M:%S")
                dict_data["trading_datetime"] = dict_data["trading_datetime"].strftime("%Y-%m-%d %H:%M:%S")
                trades_list.append(dict_data)

        # fix dataframe
        trades = pd.DataFrame(trades_list)
        if "datetime" in trades.columns:
            trades = trades.set_index("datetime")

        df = pd.DataFrame(total_portfolios_list)
        df["date"] = pd.to_datetime(df["date"])
        total_portfolios = df.set_index("date").sort_index()

        result_dict = {
            "summary": summary,
            "trades": trades,
            "total_portfolios": total_portfolios,
        }

        # plots
        if ExecutionContext.plots is not None:
            plots = ExecutionContext.plots.get_plots()
            plots_items = defaultdict(dict)
            for series_name, value_dict in six.iteritems(plots):
                for date, value in six.iteritems(value_dict):
                    plots_items[date][series_name] = value
                    plots_items[date]["date"] = date

            df = pd.DataFrame([dict_data for date, dict_data in six.iteritems(plots_items)])
            df["date"] = pd.to_datetime(df["date"])
            df = df.set_index("date").sort_index()
            result_dict["plots"] = df

        # extra
        for account_type, account in six.iteritems(self._env.accounts):
            account_name = str(account_type).split(".")[1].lower()
            portfolios_list = portfolios_list_dict[account_type]
            df = pd.DataFrame(portfolios_list)
            df["date"] = pd.to_datetime(df["date"])
            portfolios_df = df.set_index("date").sort_index()
            result_dict["{}_portfolios".format(account_name)] = portfolios_df

            positions_list = positions_list_dict[account_type]
            positions_df = pd.DataFrame(positions_list)
            if "date" in positions_df.columns:
                positions_df["date"] = pd.to_datetime(positions_df["date"])
                positions_df = positions_df.set_index("date").sort_index()
            result_dict["{}_positions".format(account_name)] = positions_df

        return result_dict

    def get_result(self):
        risk_cal = self._risk_cal
        columns = [
            "daily_returns",
            "total_returns",
            "annualized_returns",
            "market_value",
            "portfolio_value",
            # "total_commission",
            # "total_tax",
            "daily_pnl",
            "positions",
            "cash",
        ]
        risk_keys = [
            "volatility", "max_drawdown",
            "alpha", "beta", "sharpe",
            "information_ratio", "downside_risk",
            "tracking_error", "sortino",
        ]

        data = []
        for date, portfolio in six.iteritems(self._portfolios):
            # print(date, portfolio)
            # portfolio
            items = {"date": pd.Timestamp(date)}
            for key in columns:
                items[key] = getattr(portfolio, key)

            items["trades"] = self._trades[date]

            # risk
            risk = risk_cal.daily_risks[date]
            for risk_key in risk_keys:
                items[risk_key] = getattr(risk, risk_key)

            idx = risk_cal.trading_index.get_loc(date)
            items["benchmark_total_returns"] = risk_cal.benchmark_total_returns[idx]
            items["benchmark_daily_returns"] = risk_cal.benchmark_total_daily_returns[idx]
            items["benchmark_annualized_returns"] = risk_cal.benchmark_annualized_returns[idx]

            data.append(items)

        results_df = pd.DataFrame(data)
        results_df.set_index("date", inplace=True)

        return results_df



# -*- coding: utf-8 -*-
#
# Copyright 2017 Ricequant, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six
import os
import sys
from datetime import datetime
import pickle

import jsonpickle.ext.numpy as jsonpickle_numpy
import pytz

from .core.strategy import Strategy
from .api import helper as api_helper
from . import const
from .analyser.risk_cal import RiskCal
from .core.default_broker import DefaultBroker
from .core.default_event_source import DefaultEventSource
from .core.default_strategy_loader import FileStrategyLoader
from .core.strategy_universe import StrategyUniverse
from .data.base_data_source import BaseDataSource
from .data.data_proxy import DataProxy
from .environment import Environment
from .events import Events
from .execution_context import ExecutionContext
from .interface import Persistable
from .mod.mod_handler import ModHandler
from .model.bar import BarMap
from .trader.account import MixedAccount
from .trader.global_var import GlobalVars
from .trader.strategy_context import StrategyContext
from .utils import create_custom_exception, run_with_user_log_disabled
from .utils.exception import CustomException, is_user_exc, patch_user_exc
from .utils.i18n import gettext as _
from .utils.logger import user_log, system_log, user_print, user_detail_log
from .utils.persisit_helper import CoreObjectsPersistProxy, PersistHelper
from .utils.result_aggregator import ResultAggregator
from .utils.scheduler import Scheduler
from .utils import scheduler as mod_scheduler
from .plot import plot_result


jsonpickle_numpy.register_handlers()


def _adjust_start_date(config, data_proxy):
    origin_start_date, origin_end_date = config.base.start_date, config.base.end_date
    start, end = data_proxy.available_data_range(config.base.frequency)

    # print(repr(start), repr(end))
    config.base.start_date = max(start, config.base.start_date)
    config.base.end_date = min(end, config.base.end_date)
    config.base.trading_calendar = data_proxy.get_trading_dates(config.base.start_date, config.base.end_date)
    if len(config.base.trading_calendar) == 0:
        raise patch_user_exc(ValueError(_('There is no trading day between {start_date} and {end_date}.').format(
            start_date=origin_start_date, end_date=origin_end_date)))
    config.base.start_date = config.base.trading_calendar[0].date()
    config.base.end_date = config.base.trading_calendar[-1].date()
    config.base.timezone = pytz.utc


def _validate_benchmark(config, data_proxy):
    benchmark = config.base.benchmark
    if benchmark is None:
        return
    instrument = data_proxy.instruments(benchmark)
    if instrument.order_book_id == "000300.XSHG":
        # 000300.XSHG 数据进行了补齐，因此认为只要benchmark设置了000300.XSHG，就存在数据，不受限于上市日期。
        return
    if not instrument:
        raise patch_user_exc(ValueError(_('invalid benchmark {}').format(benchmark)))

    config = Environment.get_instance().config
    start_date = config.base.start_date
    end_date = config.base.end_date
    if instrument.listed_date.date() > start_date:
        raise patch_user_exc(ValueError(
            _("benchmark {benchmark} has not been listed on {start_date}").format(benchmark=benchmark,
                                                                                  start_date=start_date)))
    if instrument.de_listed_date.date() < end_date:
        raise patch_user_exc(ValueError(
            _("benchmark {benchmark} has been de_listed on {end_date}").format(benchmark=benchmark,
                                                                               end_date=end_date)))


def create_base_scope():
    import copy

    from . import user_module
    scope = copy.copy(user_module.__dict__)
    scope.update({
        "logger": user_log,
        "print": user_print,
    })

    return scope


def run(config):
    env = Environment(config)
    persist_helper = None
    init_succeed = False

    try:
        env.set_strategy_loader(FileStrategyLoader())
        env.set_global_vars(GlobalVars())
        mod_handler = ModHandler(env)
        mod_handler.start_up()

        if not env.data_source:
            env.set_data_source(BaseDataSource(config.base.data_bundle_path))

        env.set_data_proxy(DataProxy(env.data_source))
        ExecutionContext.data_proxy = env.data_proxy
        Scheduler.set_trading_dates_(env.data_source.get_trading_calendar())
        scheduler = Scheduler(config.base.frequency)
        mod_scheduler._scheduler = scheduler

        env._universe = StrategyUniverse()

        _adjust_start_date(env.config, env.data_proxy)

        _validate_benchmark(env.config, env.data_proxy)

        broker = DefaultBroker(env)
        env.accounts = accounts = broker.get_accounts()
        env.account = account = MixedAccount(accounts)

        ExecutionContext.broker = broker
        ExecutionContext.accounts = accounts
        ExecutionContext.account = account
        ExecutionContext.config = env.config

        event_source = env.event_source
        if event_source is None:
            event_source = DefaultEventSource(env.data_proxy, env.config.base.account_list)

        bar_dict = BarMap(env.data_proxy, config.base.frequency)
        ctx = ExecutionContext(const.EXECUTION_PHASE.GLOBAL, bar_dict)
        ctx._push()

        # FIXME
        start_dt = datetime.combine(config.base.start_date, datetime.min.time())
        env.calendar_dt = ExecutionContext.calendar_dt = start_dt
        env.trading_dt = ExecutionContext.trading_dt = start_dt

        env.event_bus.publish_event(Events.POST_SYSTEM_INIT)

        scope = create_base_scope()
        scope.update({
            "g": env.global_vars
        })

        apis = api_helper.get_apis(config.base.account_list)
        scope.update(apis)

        scope = env.strategy_loader.load(env.config.base.strategy_file, scope)

        if env.config.extra.enable_profiler:
            enable_profiler(env, scope)

        risk_cal = RiskCal()
        risk_cal.init(config.base.trading_calendar, is_annualized=True, save_daily_risk=True)
        result_aggregator = ResultAggregator(env, risk_cal)

        ucontext = StrategyContext()
        user_strategy = Strategy(env.event_bus, scope, ucontext)

        if not config.extra.force_run_init_when_pt_resume:
            with run_with_user_log_disabled(disabled=config.base.resume_mode):
                user_strategy.init()

        if config.extra.context_vars:
            for k, v in six.iteritems(config.extra.context_vars):
                setattr(ucontext, k, v)

        if config.base.persist:
            persist_provider = env.persist_provider
            persist_helper = PersistHelper(persist_provider, env.event_bus, config.base.persist_mode)
            persist_helper.register('core', CoreObjectsPersistProxy(scheduler))
            persist_helper.register('user_context', ucontext)
            persist_helper.register('global_vars', env.global_vars)
            persist_helper.register('universe', env._universe)
            if isinstance(event_source, Persistable):
                persist_helper.register('event_source', event_source)
            for k, v in six.iteritems(accounts):
                persist_helper.register('{}_account'.format(k.name.lower()), v)
            for name, module in six.iteritems(env.mod_dict):
                if isinstance(module, Persistable):
                    persist_helper.register('mod_{}'.format(name), module)
            # broker will restore open orders from account
            persist_helper.register('broker', broker)

            persist_helper.restore()

        init_succeed = True

        # When force_run_init_when_pt_resume is active,
        # we should run `init` after restore persist data
        if config.extra.force_run_init_when_pt_resume:
            assert config.base.resume_mode == True
            with run_with_user_log_disabled():
                user_strategy.init()

        for calendar_dt, trading_dt, event in event_source.events(config.base.start_date,
                                                                  config.base.end_date,
                                                                  config.base.frequency):
            ExecutionContext.calendar_dt = calendar_dt
            ExecutionContext.trading_dt = trading_dt
            env.calendar_dt = calendar_dt
            env.trading_dt = trading_dt
            for account in accounts.values():
                account.portfolio._current_date = trading_dt.date()

            if event == Events.BEFORE_TRADING:
                env.event_bus.publish_event(Events.PRE_BEFORE_TRADING)
                broker.before_trading()
                for account in accounts.values():
                    account.before_trading()
                scheduler.next_day_(trading_dt)
                env.event_bus.publish_event(Events.BEFORE_TRADING)
                with ExecutionContext(const.EXECUTION_PHASE.BEFORE_TRADING):
                    scheduler.before_trading_(ucontext)
                env.event_bus.publish_event(Events.POST_BEFORE_TRADING)
            elif event == Events.BAR:
                bar_dict.update_dt(calendar_dt)
                broker.update(calendar_dt, trading_dt, bar_dict)
                for account in accounts.values():
                    account.on_bar(bar_dict)
                env.event_bus.publish_event(Events.PRE_BAR)
                env.event_bus.publish_event(Events.BAR, bar_dict)
                with ExecutionContext(const.EXECUTION_PHASE.SCHEDULED, bar_dict):
                    scheduler.next_bar_(ucontext, bar_dict)
                env.event_bus.publish_event(Events.POST_BAR)
            elif event == Events.TICK:
                env.event_bus.publish_event(Events.PRE_TICK)
                env.event_bus.publish_event(Events.TICK)
                env.event_bus.publish_event(Events.POST_TICK)
            elif event == Events.AFTER_TRADING:
                env.event_bus.publish_event(Events.PRE_AFTER_TRADING)
                broker.after_trading()
                for account in accounts.values():
                    account.after_trading()
                env.event_bus.publish_event(Events.AFTER_TRADING)
                env.event_bus.publish_event(Events.POST_AFTER_TRADING)
            elif event == Events.SETTLEMENT:
                env.event_bus.publish_event(Events.PRE_SETTLEMENT)
                for account in accounts.values():
                    account.settlement()
                env.event_bus.publish_event(Events.SETTLEMENT)
                env.event_bus.publish_event(Events.POST_SETTLEMENT)
            else:
                raise RuntimeError('unknown event from event source: {}'.format(event))

        strategy_name = os.path.basename(env.config.base.strategy_file).split(".")[0]
        result_dict = result_aggregator.get_result_dict(strategy_name)

        output_generated_results(env, result_dict)

        mod_handler.tear_down(const.EXIT_CODE.EXIT_SUCCESS)
        system_log.debug("strategy run successfully, normal exit")
        return result_dict
    except CustomException as e:
        if init_succeed and env.config.base.persist and persist_helper:
            persist_helper.persist()

        user_detail_log.exception("strategy execute exception")
        user_log.error(e.error)
        mod_handler.tear_down(const.EXIT_CODE.EXIT_USER_ERROR, e)
    except Exception as e:
        if init_succeed and env.config.base.persist and persist_helper:
            persist_helper.persist()

        exc_type, exc_val, exc_tb = sys.exc_info()
        user_exc = create_custom_exception(exc_type, exc_val, exc_tb, config.base.strategy_file)

        user_log.error(user_exc.error)
        code = const.EXIT_CODE.EXIT_USER_ERROR
        if not is_user_exc(exc_val):
            system_log.exception("strategy execute exception")
            code = const.EXIT_CODE.EXIT_INTERNAL_ERROR
        else:
            user_detail_log.exception("strategy execute exception")

        mod_handler.tear_down(code, user_exc)


def enable_profiler(env, scope):
    # decorate line profiler
    import line_profiler
    import inspect
    env.profile_deco = profile_deco = line_profiler.LineProfiler()
    for name in scope:
        obj = scope[name]
        if getattr(obj, "__module__", None) != "rqalpha.user_module":
            continue
        if inspect.isfunction(obj):
            scope[name] = profile_deco(obj)
        if inspect.isclass(obj):
            for key, val in six.iteritems(obj.__dict__):
                if inspect.isfunction(val):
                    setattr(obj, key, profile_deco(val))


def output_generated_results(env, result_dict):
    # output pickle file
    if env.config.extra.output_file is not None:
        pickle.dump(result_dict, open(env.config.extra.output_file, "wb"))

    # plot
    if env.config.extra.plot:
        plot_result(result_dict, show_windows=env.config.extra.plot)

    # save plot as image file
    if env.config.extra.plot_save_file:
        plot_result(result_dict, show_windows=False,
                    savefile=env.config.extra.plot_save_file)

    # generate report
    if env.config.extra.report_save_path:
        generate_report(result_dict, env.config.extra.report_save_path)

    # output line profiler result
    if env.profile_deco:  # config.extra.enable_profiler:
        from six import StringIO
        # profile_deco.dump_stats("{}.lprof".format(strategy_name))
        stdout_trap = StringIO()
        env.profile_deco.print_stats(stdout_trap)
        profile_output = stdout_trap.getvalue()
        profile_output = profile_output.rstrip()
        print(profile_output)
        env.event_bus.publish_event(Events.ON_LINE_PROFILER_RESULT, profile_output)


def generate_report(result_dict, target_report_csv_path):
    import pandas as pd
    from six import StringIO

    output_path = os.path.join(target_report_csv_path, result_dict["summary"]["strategy_name"])
    try:
        os.mkdir(output_path)
    except:
        pass

    xlsx_writer = pd.ExcelWriter(os.path.join(output_path, "report.xlsx"), engine='xlsxwriter')

    # summary.csv
    csv_txt = StringIO()
    summary = result_dict["summary"]
    csv_txt.write(u"\n".join(sorted("{},{}".format(key, value) for key, value in six.iteritems(summary))))
    df = pd.DataFrame(data=[{"val": val} for val in summary.values()], index=summary.keys()).sort_index()
    df.to_excel(xlsx_writer, sheet_name="summary")

    with open(os.path.join(output_path, "summary.csv"), 'w') as csvfile:
        csvfile.write(csv_txt.getvalue())

    for name in ["total_portfolios", "stock_portfolios", "future_portfolios",
                 "stock_positions", "future_positions", "trades"]:
        try:
            df = result_dict[name]
        except KeyError:
            continue

        # replace all date in dataframe as string
        if df.index.name == "date":
            df = df.reset_index()
            df["date"] = df["date"].apply(lambda x: x.strftime("%Y-%m-%d"))
            df = df.set_index("date")

        csv_txt = StringIO()
        csv_txt.write(df.to_csv(encoding='utf-8'))

        df.to_excel(xlsx_writer, sheet_name=name)

        with open(os.path.join(output_path, "{}.csv".format(name)), 'w') as csvfile:
            csvfile.write(csv_txt.getvalue())

    # report.xls <--- 所有sheet的汇总
    xlsx_writer.save()



# -*- coding: utf-8 -*-
#
# Copyright 2017 Ricequant, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division
import six
from collections import OrderedDict
import numpy as np
import pandas as pd
import jsonpickle

from .risk import Risk
from .. import const
from ..utils.logger import system_log


class RiskCal(object):
    def __init__(self):
        self.initialized = False

    def init(self, trading_index, is_annualized=True, save_daily_risk=True):
        self.is_annualized = is_annualized
        self.save_daily_risk = save_daily_risk

        self.date = self.start_date = None
        self.trading_index = trading_index

        self.date = self.start_date = self.trading_index[0].date()
        trading_days_cnt = len(self.trading_index)

        self.strategy_total_daily_returns = np.full(trading_days_cnt, np.nan)
        self.strategy_total_returns = np.full(trading_days_cnt, np.nan)
        self.strategy_annualized_returns = np.full(trading_days_cnt, np.nan)
        self.benchmark_total_daily_returns = np.full(trading_days_cnt, np.nan)
        self.benchmark_total_returns = np.full(trading_days_cnt, np.nan)
        self.benchmark_annualized_returns = np.full(trading_days_cnt, np.nan)

        self.strategy_current_daily_returns = self.strategy_total_daily_returns[:0]
        self.strategy_current_total_returns = self.strategy_total_returns[:0]
        self.strategy_current_annualized_returns = self.strategy_annualized_returns[:0]
        self.benchmark_current_daily_returns = self.benchmark_total_daily_returns[:0]
        self.benchmark_current_total_returns = self.benchmark_total_returns[:0]
        self.benchmark_current_annualized_returns = self.benchmark_annualized_returns[:0]

        self._risk = Risk()
        self.daily_risks = OrderedDict()

        self.current_max_returns = -np.inf
        self.current_max_drawdown = 0
        self.initialized = True

    def update_trading_index(self, trading_index):
        """PT的交易日可能会随着时间变长
        """
        if len(self.trading_index) == len(trading_index):
            return

        self.trading_index = trading_index
        size = len(self.trading_index)

        attrs = [
            "strategy_total_daily_returns", "benchmark_total_daily_returns",
            "strategy_total_returns", "benchmark_total_returns",
            "strategy_annualized_returns", "benchmark_annualized_returns",
        ]
        for attr in attrs:
            setattr(self, attr, np.resize(getattr(self, attr), (size, )))

    def calculate(self, date, strategy_daily_returns, benchmark_daily_returns, risk_free_rate):
        """calculate daily risk

        :param datetime.date date: current date
        :param float strategy_daily_returns: strategy daily return
        :param float benchmark_daily_returns: benchmark daily return
        :param float risk_free_rate: risk free period return
        :returns: today risk
        :rtype: Risk
        """
        try:
            idx = self.latest_idx = self.trading_index.get_loc(date)
        except KeyError as e:
            system_log.exception("trading_index.get_loc({}) fail", date)
            return

        self.date = date

        # daily
        self.strategy_total_daily_returns[idx] = strategy_daily_returns
        self.benchmark_total_daily_returns[idx] = benchmark_daily_returns
        self.strategy_current_daily_returns = self.strategy_total_daily_returns[:idx + 1]
        self.benchmark_current_daily_returns = self.benchmark_total_daily_returns[:idx + 1]

        self.days_cnt = len(self.strategy_current_daily_returns)
        days_pass_cnt = (date - self.start_date).days + 1

        # risk
        self.riskfree_total_returns = risk_free_rate

        # total
        self.strategy_total_returns[idx] = (1. + self.strategy_current_daily_returns).prod() - 1
        self.benchmark_total_returns[idx] = (1. + self.benchmark_current_daily_returns).prod() - 1
        self.strategy_current_total_returns = self.strategy_total_returns[:idx + 1]
        self.benchmark_current_total_returns = self.benchmark_total_returns[:idx + 1]

        # annual
        self.strategy_annualized_returns[idx] = (1 + self.strategy_current_total_returns[-1]) ** (
                    const.DAYS_CNT.DAYS_A_YEAR / days_pass_cnt) - 1
        self.benchmark_annualized_returns[idx] = (1 + self.benchmark_current_total_returns[-1]) ** (
            const.DAYS_CNT.DAYS_A_YEAR / days_pass_cnt) - 1
        self.strategy_current_annualized_returns = self.strategy_annualized_returns[:idx + 1]
        self.benchmark_current_annualized_returns = self.benchmark_annualized_returns[:idx + 1]

        if self.strategy_current_total_returns[-1] > self.current_max_returns:
            self.current_max_returns = self.strategy_current_total_returns[-1]

        # need to update daily
        self._risk.max_drawdown = self.cal_max_drawdown()

    @property
    def risk(self):
        risk = self._risk

        try:
            risk.volatility = self.cal_volatility()
            # risk.max_drawdown = self.cal_max_drawdown()
            risk.tracking_error = self.cal_tracking_error()
            risk.information_ratio = self.cal_information_ratio(risk.tracking_error)
            risk.downside_risk = self.cal_downside_risk()
            risk.beta = self.cal_beta()
            risk.alpha = self.cal_alpha()
            risk.sharpe = self.cal_sharpe()
            risk.sortino = self.cal_sortino()
        except Exception as e:
            system_log.exception("risk cal exc")

        if self.save_daily_risk:
            self.daily_risks[self.date] = risk._clone()
            return self.daily_risks[self.date]
        else:
            return risk._clone()

    def cal_volatility(self):
        daily_returns = self.strategy_current_daily_returns
        if len(daily_returns) <= 1:
            return 0.

        if self.is_annualized:
            volatility = const.DAYS_CNT.TRADING_DAYS_A_YEAR ** 0.5 * np.std(daily_returns, ddof=1)
        else:
            volatility = len(daily_returns) ** 0.5 * np.std(daily_returns, ddof=1)

        return volatility

    def cal_max_drawdown(self):
        today_return = self.strategy_current_total_returns[-1]
        today_drawdown = (1. + today_return) / (1. + self.current_max_returns) - 1.
        if today_drawdown < self.current_max_drawdown:
            self.current_max_drawdown = today_drawdown
        return abs(self.current_max_drawdown)

    def cal_tracking_error(self):
        diff = self.strategy_current_daily_returns - self.benchmark_current_daily_returns

        if self.is_annualized:
            return ((diff * diff).sum() / len(diff)) ** 0.5 * const.DAYS_CNT.TRADING_DAYS_A_YEAR ** 0.5
        else:
            return ((diff * diff).sum() / len(diff)) ** 0.5 * len(diff) ** 0.5

    def cal_information_ratio(self, volatility):
        strategy_rets = self.strategy_current_annualized_returns[-1]
        benchmark_rets = self.benchmark_current_annualized_returns[-1]

        if volatility == 0:
            return np.nan
        return (strategy_rets - benchmark_rets) / volatility

    def cal_alpha(self):
        beta = self._risk.beta

        strategy_rets = self.strategy_current_annualized_returns[-1]
        benchmark_rets = self.benchmark_current_annualized_returns[-1]

        alpha = strategy_rets - (self.riskfree_total_returns + beta * (benchmark_rets - self.riskfree_total_returns))
        return alpha

    def cal_beta(self):
        if len(self.strategy_current_daily_returns) <= 1:
            return 0.
        cov = np.cov(np.vstack([
            self.strategy_current_daily_returns,
            self.benchmark_current_daily_returns,
        ]), ddof=1)
        val = cov[1][1]
        if val == 0:
            return np.nan
        beta = cov[0][1] / val

        return beta

    def cal_sharpe(self):
        volatility = self._risk.volatility
        strategy_rets = self.strategy_current_annualized_returns[-1]

        if volatility != 0:
            sharpe = (strategy_rets - self.riskfree_total_returns) / volatility
        else:
            sharpe = np.nan

        return sharpe

    def cal_sortino(self):
        strategy_rets = self.strategy_current_annualized_returns[-1]
        downside_risk = self._risk.downside_risk

        if downside_risk != 0:
            sortino = (strategy_rets - self.riskfree_total_returns) / downside_risk
        else:
            sortino = np.nan
        return sortino

    def cal_downside_risk(self):
        mask = self.strategy_current_daily_returns < self.benchmark_current_daily_returns
        diff = self.strategy_current_daily_returns[mask] - self.benchmark_current_daily_returns[mask]
        if len(diff) <= 1:
            return 0.

        if self.is_annualized:
            return ((diff * diff).sum() / len(diff)) ** 0.5 * const.DAYS_CNT.TRADING_DAYS_A_YEAR ** 0.5
        else:
            return ((diff * diff).sum() / len(diff)) ** 0.5 * len(diff) ** 0.5

    def __repr__(self):
        return "RiskCal({0})".format(self.__dict__)

    def to_json(self):
        blacklist = ["trading_index"]
        data = {key: value for key, value in six.iteritems(self.__dict__) if key not in blacklist}
        data["trading_index"] = self.trading_index.tolist()
        json_str = jsonpickle.encode(data)
        return json_str

    def from_json(self, json_str):
        data = jsonpickle.decode(json_str)
        data["trading_index"] = pd.Index(data["trading_index"])

        self.__dict__.update(data)

    def get_state(self):

        return self.to_json().encode('utf-8')

    def set_state(self, state):
        self.from_json(state.decode('utf-8'))




if __name__ == '__main__':
    import os
    import datetime
    from rqalpha.data import LocalDataProxy

    data_proxy = LocalDataProxy(os.path.expanduser("~/data/bundle"))
    trading_dates = data_proxy.get_trading_dates("2016-11-01", "2050-01-01")

    risk_cal = RiskCal(None)
    risk_cal.init_data(trading_dates)

    risk_cal.calculate(datetime.date(2016, 11, 1), 0.01, 0.01, 0.03)

    risk_cal2 = RiskCal(None)

    risk_cal2.from_json(risk_cal.to_json())



# -*- coding: utf-8 -*-
#
# Copyright 2017 Ricequant, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .future_portfolio import FuturePortfolio
from .stock_portfolio import StockPortfolio
from .base_portfolio import BasePortfolio
from ...const import ACCOUNT_TYPE


def init_portfolio(init_cash, start_date, account_type):
    if account_type == ACCOUNT_TYPE.STOCK:
        return StockPortfolio(init_cash, start_date, ACCOUNT_TYPE.STOCK)
    elif account_type == ACCOUNT_TYPE.FUTURE:
        return FuturePortfolio(init_cash, start_date, ACCOUNT_TYPE.FUTURE)
    elif account_type == ACCOUNT_TYPE.BENCHMARK:
        return StockPortfolio(init_cash, start_date, ACCOUNT_TYPE.BENCHMARK)


# -*- coding: utf-8 -*-
#
# Copyright 2017 Ricequant, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click
import datetime
import errno
import locale
import os
import shutil
import tarfile
import tempfile

import six
import requests
import pandas as pd

from .cache_control import set_cache_policy, CachePolicy
from .utils.click_helper import Date
from .utils.i18n import set_locale
from .utils.config import parse_config


@click.group()
@click.option('-v', '--verbose', count=True)
@click.pass_context
def cli(ctx, verbose):
    ctx.obj["VERBOSE"] = verbose


def entry_point():
    cli(obj={})


@cli.command()
@click.option('-d', '--data-bundle-path', default=os.path.expanduser("~/.rqalpha"), type=click.Path())
def update_bundle(data_bundle_path):
    """update data bundle, download if not found"""
    day = datetime.date.today()
    tmp = os.path.join(tempfile.gettempdir(), 'rq.bundle')

    while True:
        url = 'http://7xjci3.com1.z0.glb.clouddn.com/bundles_v2/rqbundle_%04d%02d%02d.tar.bz2' % (
            day.year, day.month, day.day)
        six.print_('try {} ...'.format(url))
        r = requests.get(url, stream=True)
        if r.status_code != 200:
            day = day - datetime.timedelta(days=1)
            continue

        out = open(tmp, 'wb')
        total_length = int(r.headers.get('content-length'))

        with click.progressbar(length=total_length, label='downloading ...') as bar:
            for data in r.iter_content(chunk_size=8192):
                bar.update(len(data))
                out.write(data)

        out.close()
        break

    shutil.rmtree(data_bundle_path, ignore_errors=True)
    os.mkdir(data_bundle_path)
    tar = tarfile.open(tmp, 'r:bz2')
    tar.extractall(data_bundle_path)
    tar.close()
    os.remove(tmp)


@cli.command()
@click.option('-d', '--data-bundle-path', 'base__data_bundle_path', type=click.Path(exists=True))
@click.option('-f', '--strategy-file', 'base__strategy_file', type=click.Path(exists=True))
@click.option('-s', '--start-date', 'base__start_date', type=Date())
@click.option('-e', '--end-date', 'base__end_date', type=Date())
@click.option('-r', '--rid', 'base__run_id', type=click.STRING)
@click.option('-i', '--init-cash', 'base__stock_starting_cash', type=click.FLOAT)
@click.option('-sc', '--stock-starting-cash', 'base__stock_starting_cash', type=click.FLOAT)
@click.option('-fc', '--future-starting-cash', 'base__future_starting_cash', type=click.FLOAT)
@click.option('-bm', '--benchmark', 'base__benchmark', type=click.STRING, default=None)
@click.option('-sp', '--slippage', 'base__slippage', type=click.FLOAT)
@click.option('-cm', '--commission-multiplier', 'base__commission_multiplier', type=click.FLOAT)
@click.option('-mm', '--margin-multiplier', 'base__margin_multiplier', type=click.FLOAT)
@click.option('-k', '--kind', 'base__strategy_type', type=click.Choice(['stock', 'future', 'stock_future']))
@click.option('-fq', '--frequency', 'base__frequency', type=click.Choice(['1d', '1m']))
@click.option('-me', '--match-engine', 'base__matching_type', type=click.Choice(['current_bar', 'next_bar']))
@click.option('-rt', '--run-type', 'base__run_type', type=click.Choice(['b', 'p']), default="b")
@click.option('--resume', 'base__resume_mode', is_flag=True)
@click.option('--handle-split/--not-handle-split', 'base__handle_split', default=None, help="handle split")
@click.option('--risk-grid/--no-risk-grid', 'base__cal_risk_grid', default=True)
@click.option('-l', '--log-level', 'extra__log_level', type=click.Choice(['verbose', 'debug', 'info', 'error']))
@click.option('-p', '--plot/--no-plot', 'extra__plot', default=None, help="plot result")
@click.option('-o', '--output-file', 'extra__output_file', type=click.Path(writable=True), help="output result pickle file")
@click.option('--fast-match', 'validator__fast_match', is_flag=True)
@click.option('--progress/--no-progress', 'mod__progress__enabled', default=None, help="show progress bar")
@click.option('--extra-vars', 'extra__context_vars', type=click.STRING, help="override context vars")
@click.option("--enable-profiler", "extra__enable_profiler", is_flag=True, help="add line profiler to profile your strategy")
@click.option('--config', 'config_path', type=click.STRING, help="config file path")
@click.help_option('-h', '--help')
def run(**kwargs):
    if kwargs.get('base__run_type') == 'p':
        set_cache_policy(CachePolicy.MINIMUM)

    try:
        locale.setlocale(locale.LC_ALL, "en_US.UTF-8")
        locale.setlocale(locale.LC_CTYPE, "en_US.UTF-8")
        os.environ['TZ'] = 'Asia/Shanghai'
        set_locale(["zh_Hans_CN"])
    except Exception as e:
        if os.name != 'nt':
            raise

    config_path = kwargs.get('config_path', None)
    if config_path is not None:
        config_path = os.path.abspath(config_path)

    from . import main
    main.run(parse_config(kwargs, config_path))


@cli.command()
@click.option('-d', '--directory', default="./", type=click.Path(), required=True)
def examples(directory):
    """
    generate example strategies to target folder
    """
    source_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), "examples")

    try:
        shutil.copytree(source_dir, os.path.join(directory, "examples"))
    except OSError as e:
        if e.errno == errno.EEXIST:
            print("Folder examples is exists.")


@cli.command()
@click.argument('result_dict_file', type=click.Path(exists=True), required=True)
@click.option('--show/--hide', 'is_show', default=True)
@click.option('--plot-save', 'plot_save_file', default=None, type=click.Path(), help="save plot result to file")
def plot(result_dict_file, is_show, plot_save_file):
    """
    draw result DataFrame
    """
    import pandas as pd
    from rqalpha.plot import plot_result

    result_dict = pd.read_pickle(result_dict_file)
    if is_show:
        plot_result(result_dict)
    if plot_save_file:
        plot_result(result_dict, show_windows=False, savefile=plot_save_file)


@cli.command()
@click.argument('result_pickle_file_path', type=click.Path(exists=True), required=True)
@click.argument('target_report_csv_path', type=click.Path(exists=True, writable=True), required=True)
def report(result_pickle_file_path, target_report_csv_path):
    """
    generate report from backtest output file
    """
    import pandas as pd
    result_dict = pd.read_pickle(result_pickle_file_path)

    from rqalpha import main

    main.generate_report(result_dict, target_report_csv_path)


@cli.command()
@click.option('-v', '--verbose', is_flag=True)
def version(**kwargs):
    """
    Output Version Info
    """
    from rqalpha import version_info
    print("Current Version: ", version_info)


if __name__ == '__main__':
    entry_point()



# -*- coding: utf-8 -*-
#
# Copyright 2017 Ricequant, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..utils.repr import dict_repr


class Risk(object):

    __repr__ = dict_repr

    def __init__(self):
        self.volatility = .0
        self.alpha = .0
        self.beta = .0
        self.sharpe = .0
        self.sortino = .0
        self.information_ratio = .0
        self.max_drawdown = .0
        self.tracking_error = .0
        self.downside_risk = .0

    def _clone(self):
        c = type('RiskClone', (), {})()
        c.volatility = self.volatility
        c.alpha = self.alpha
        c.beta = self.beta
        c.sharpe = self.sharpe
        c.sortino = self.sortino
        c.information_ratio = self.information_ratio
        c.max_drawdown = self.max_drawdown
        c.tracking_error = self.tracking_error
        c.downside_risk = self.downside_risk
        return c

    def __to_dict__(self):
        risk_dict = {
            'volatility': self.volatility,
            'alpha': self.alpha,
            'beta': self.beta,
            'sharpe': self.sharpe,
            'sortino': self.sortino,
            'information_ratio': self.information_ratio,
            'max_drawdown': self.max_drawdown,
            'tracking_error': self.tracking_error,
            'downside_risk': self.downside_risk
        }
        return risk_dict

    @classmethod
    def __from_dict__(cls, risk_dict):
        risk = cls()
        risk.volatility = risk_dict['volatility']
        risk.alpha = risk_dict['alpha']
        risk.beta = risk_dict['beta']
        risk.sharpe = risk_dict['sharpe']
        risk.sortino = risk_dict['sortino']
        risk.information_ratio = risk_dict['information_ratio']
        risk.max_drawdown = risk_dict['max_drawdown']
        risk.tracking_error = risk_dict['tracking_error']
        risk.downside_risk = risk_dict['downside_risk']
        return risk



# -*- coding: utf-8 -*-
#
# Copyright 2017 Ricequant, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six
import os
import yaml
import datetime
import logbook
from pprint import pformat
import codecs

from . import RqAttrDict, logger
from .exception import patch_user_exc
from .logger import user_log, system_log, std_log, user_std_handler
from ..const import ACCOUNT_TYPE, MATCHING_TYPE, RUN_TYPE, PERSIST_MODE
from ..utils.i18n import gettext as _


def parse_config(click_kargs, base_config_path=None):
    if base_config_path is None:
        config_path = os.path.abspath(os.path.join(os.path.dirname(os.path.realpath(__file__)), "../config.yml"))
    else:
        config_path = base_config_path
    if not os.path.exists(config_path):
        print("config.yaml not found in", os.getcwd())
        return

    with codecs.open(config_path, encoding="utf-8") as f:
        config_file = f.read()

    config = yaml.load(config_file)
    for key, value in six.iteritems(click_kargs):
        if click_kargs[key] is not None:
            keys = key.split("__")
            keys.reverse()
            sub_config = config[keys.pop()]
            while True:
                if len(keys) == 0:
                    break
                k = keys.pop()
                if len(keys) == 0:
                    sub_config[k] = value
                else:
                    sub_config = sub_config[k]

    config = parse_user_config(config)

    config = RqAttrDict(config)
    base_config = config.base
    extra_config = config.extra

    if isinstance(base_config.start_date, six.string_types):
        base_config.start_date = datetime.datetime.strptime(base_config.start_date, "%Y-%m-%d")
    if isinstance(base_config.start_date, datetime.datetime):
        base_config.start_date = base_config.start_date.date()
    if isinstance(base_config.end_date, six.string_types):
        base_config.end_date = datetime.datetime.strptime(base_config.end_date, "%Y-%m-%d")
    if isinstance(base_config.end_date, datetime.datetime):
        base_config.end_date = base_config.end_date.date()
    if base_config.commission_multiplier < 0:
        raise patch_user_exc(ValueError(_("invalid commission multiplier value: value range is [0, +∞)")))
    if base_config.margin_multiplier <= 0:
        raise patch_user_exc(ValueError(_("invalid margin multiplier value: value range is (0, +∞]")))

    if base_config.data_bundle_path is None:
        base_config.data_bundle_path = os.path.expanduser("~/.rqalpha")

    if not os.path.exists(base_config.data_bundle_path):
        print("data bundle not found. Run `rqalpha update_bundle` to download data bundle.")
        # print("data bundle not found. Run `%s update_bundle` to download data bundle." % sys.argv[0])
        return

    if not os.path.exists(base_config.strategy_file):
        print("strategy file not found: ", base_config.strategy_file)
        return

    base_config.run_type = parse_run_type(base_config.run_type)
    base_config.account_list = gen_account_list(base_config.strategy_type)
    base_config.matching_type = parse_matching_type(base_config.matching_type)
    base_config.persist_mode = parse_persist_mode(base_config.persist_mode)

    if extra_config.log_level == "verbose":
        user_log.handlers.append(user_std_handler)
    if extra_config.context_vars:
        import base64
        import json
        extra_config.context_vars = json.loads(base64.b64decode(extra_config.context_vars).decode('utf-8'))

    if base_config.stock_starting_cash < 0:
        raise patch_user_exc(ValueError(_('invalid stock starting cash: {}').format(base_config.stock_starting_cash)))

    if base_config.future_starting_cash < 0:
        raise patch_user_exc(ValueError(_('invalid future starting cash: {}').format(base_config.future_starting_cash)))

    if base_config.stock_starting_cash + base_config.future_starting_cash == 0:
        raise patch_user_exc(ValueError(_('stock starting cash and future starting cash can not be both 0.')))

    system_log.level = getattr(logbook, extra_config.log_level.upper(), logbook.NOTSET)
    std_log.level = getattr(logbook, extra_config.log_level.upper(), logbook.NOTSET)

    if base_config.frequency == "1d":
        logger.DATETIME_FORMAT = "%Y-%m-%d"
        config.validator.fast_match = True

    system_log.debug(pformat(config))

    return config


def parse_user_config(config):
    try:
        with codecs.open(config["base"]["strategy_file"], encoding="utf-8") as f:
            source_code = f.read()

        scope = {}

        code = compile(source_code, config["base"]["strategy_file"], 'exec')
        six.exec_(code, scope)

        __config__ = scope.get("__config__", {})

        for sub_key, sub_dict in six.iteritems(__config__):
            if sub_key not in config["whitelist"]:
                continue
            for key, value in six.iteritems(sub_dict):
                if isinstance(config[sub_key][key], dict):
                    config[sub_key][key].update(value)
                else:
                    config[sub_key][key] = value

    except Exception as e:
        print('in parse_user_config, exception: ', e)
    finally:
        return config


def gen_account_list(account_list_str):
    assert isinstance(account_list_str, six.string_types)
    account_list = account_list_str.split("_")
    return [parse_account_type(account_str) for account_str in account_list]


def parse_account_type(account_type_str):
    assert isinstance(account_type_str, six.string_types)
    if account_type_str == "stock":
        return ACCOUNT_TYPE.STOCK
    elif account_type_str == "future":
        return ACCOUNT_TYPE.FUTURE


def parse_matching_type(me_str):
    assert isinstance(me_str, six.string_types)
    if me_str == "current_bar":
        return MATCHING_TYPE.CURRENT_BAR_CLOSE
    elif me_str == "next_bar":
        return MATCHING_TYPE.NEXT_BAR_OPEN
    else:
        raise NotImplementedError


def parse_run_type(rt_str):
    assert isinstance(rt_str, six.string_types)
    if rt_str == "b":
        return RUN_TYPE.BACKTEST
    elif rt_str == "p":
        return RUN_TYPE.PAPER_TRADING
    elif rt_str == "CTP":
        return RUN_TYPE.CTP
    else:
        raise NotImplementedError


def parse_persist_mode(persist_mode):
    assert isinstance(persist_mode, six.string_types)
    if persist_mode == 'real_time':
        return PERSIST_MODE.REAL_TIME
    elif persist_mode == 'on_crash':
        return PERSIST_MODE.ON_CRASH
    else:
        raise RuntimeError('unknown persist mode: {}'.format(persist_mode))




# -*- coding: utf-8 -*-
#
# Copyright 2017 Ricequant, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc


from six import with_metaclass


class AbstractStrategyLoader(with_metaclass(abc.ABCMeta)):
    """
    策略加载器，其主要作用是加载策略，并将策略运行所需要的域环境传递给策略执行代码。
    """

    @abc.abstractmethod
    def load(self, strategy, scope):
        """
        【Required】

        load 函数负责组装策略代码和策略代码所在的域，并输出最终组装好的可执行域

        :param str strategy: 策略相关参数，可以是路径，也可以是源码本身，
            原则就是根据strategy可以获取到源码，
            系统会将 `config.base.strategy_file` 作为参数传递给load函数。

        :param dict scope: 域，系统会将默认的API函数等内容加载到scope中，这样函数就可以直接调用对应的API了，
            通过自定义load函数，可以更改scope，来增加自定义API 等功能。

        :return: scope
        """
        raise NotImplementedError


class AbstractEventSource(object):
    """
    事件生成模块，RQAlpha 会使用该模块来实现完整的事件触发
    """
    def events(self, start_date, end_date, frequency):
        """
        【Required】

        扩展 EventSource 必须实现 events 函数。

        events 是一个 event generator, 在相应事件的时候需要以如下格式来传递事件

        .. code-block:: python

            yield trading_datetime, calendar_datetime, EventEnum

        其中 trading_datetime 为基于交易日的 datetime, calendar_datetime 为基于自然日的 datetime (因为夜盘的存在，交易日和自然日未必相同)

        EventEnum 为 :class:`~Events`

        :param datetime.date start_date: 起始日期, 系统会将 `config.base.start_date` 传递 events 函数

        :param datetime.date end_date: 结束日期，系统会将 `config.base.end_date` 传递给 events 函数

        :param str frequency: 周期频率，`1d` 表示日周期, `1m` 表示分钟周期

        :return: None
        """
        raise NotImplementedError


class AbstractDataSource(object):
    """
    数据源模块，RQAlpha 支持自定义数据源，也可以基于已有默认数据源模块进行扩展，
    因为 RQAlpha 内部使用了一些数据，所以必须实现对应的接口函数。
    """
    def get_all_instruments(self):
        """
        获取所有Instrument。

        :return: list[:class:`~StockInstrument` | :class:`~FutureInstrument`]
        """
        raise NotImplementedError

    def get_trading_calendar(self):
        """
        获取交易日历

        :return: list[`pandas.Timestamp`]
        """
        raise NotImplementedError

    def get_yield_curve(self, start_date, end_date, tenor=None):
        """
        获取国债利率

        :param pandas.Timestamp str start_date: 开始日期

        :param pandas.Timestamp end_date: 结束日期

        :param bool tenor: 利率期限

        :return: pandas.DataFrame, [start_date, end_date]
        """
        raise NotImplementedError

    def get_dividend(self, order_book_id, adjusted=True):
        """
        获取股票/基金分红信息

        :param str order_book_id: 合约名
        :param bool adjusted: 是否经过前复权处理
        :return:
        """
        raise NotImplementedError

    def get_split(self, order_book_id):
        """
        获取拆股信息

        :param str order_book_id: 合约名

        :return: `pandas.DataFrame`
        """

        raise NotImplementedError

    def get_bar(self, instrument, dt, frequency):
        """
        根据 dt 来获取 对应的Bar 数据

        :param instrument: 合约对象
        :type instrument: :class:`~StockInstrument` | :class:`~FutureInstrument`

        :param datetime.datetime dt: calendar_datetime

        :param str frequency: 周期频率，`1d` 表示日周期, `1m` 表示分钟周期

        :return: `numpy.ndarray` | `dict`
        """
        raise NotImplementedError

    def get_settle_price(self, instrument, date):
        """
        获取期货品种在date的结算价

        :param instrument: 合约对象
        :type instrument: :class:`~StockInstrument` | :class:`~FutureInstrument`

        :param datetime.date date: 结算日期

        :return: `str`
        """
        raise NotImplementedError

    def history_bars(self, instrument, bar_count, frequency, fields, dt, skip_suspended=True):
        """
        获取历史数据

        :param instrument: 合约对象
        :type instrument: :class:`~StockInstrument` | :class:`~FutureInstrument`

        :param int bar_count: 获取的历史数据数量

        :param str frequency: 周期频率，`1d` 表示日周期, `1m` 表示分钟周期

        :param str fields: 返回数据字段

        =========================   ===================================================
        fields                      字段名
        =========================   ===================================================
        datetime                    时间戳
        open                        开盘价
        high                        最高价
        low                         最低价
        close                       收盘价
        volume                      成交量
        total_turnover              成交额
        datetime                    int类型时间戳
        open_interest               持仓量（期货专用）
        basis_spread                期现差（股指期货专用）
        settlement                  结算价（期货日线专用）
        prev_settlement             结算价（期货日线专用）
        =========================   ===================================================

        :param datetime.datetime dt: 时间

        :param bool skip_suspended: 是否跳过停牌日

        :return: `numpy.ndarray`

        """
        raise NotImplementedError

    def current_snapshot(self, instrument, frequency, dt):
        """
        获得当前市场快照数据。只能在日内交易阶段调用，获取当日调用时点的市场快照数据。
        市场快照数据记录了每日从开盘到当前的数据信息，可以理解为一个动态的day bar数据。
        在目前分钟回测中，快照数据为当日所有分钟线累积而成，一般情况下，最后一个分钟线获取到的快照数据应当与当日的日线行情保持一致。
        需要注意，在实盘模拟中，该函数返回的是调用当时的市场快照情况，所以在同一个handle_bar中不同时点调用可能返回的数据不同。
        如果当日截止到调用时候对应股票没有任何成交，那么snapshot中的close, high, low, last几个价格水平都将以0表示。

        :param instrument: 合约对象
        :type instrument: :class:`~StockInstrument` | :class:`~FutureInstrument`

        :param str frequency: 周期频率，`1d` 表示日周期, `1m` 表示分钟周期

        :param datetime.datetime dt: 时间

        :return: :class:`~Snapshot`
        """
        raise NotImplementedError

    def get_trading_minutes_for(self, instrument, trading_dt):
        """

        获取证券某天的交易时段，用于期货回测

        :param instrument: 合约对象
        :type instrument: :class:`~StockInstrument` | :class:`~FutureInstrument`

        :param datetime.datetime trading_dt: 交易日。注意期货夜盘所属交易日规则。

        :return: list[`datetime.datetime`]
        """
        raise NotImplementedError

    def available_data_range(self, frequency):
        """
        此数据源能提供数据的时间范围

        :param str frequency: 周期频率，`1d` 表示日周期, `1m` 表示分钟周期

        :return: (earliest, latest)
        """
        raise NotImplementedError


class AbstractBroker(object):
    """
    RQAlpha 中将 Broker 作为虚拟券商的一个代理。

    比如作为回测，通过AbstractBroker扩展了default_broker(相当于BackTestBroker)
    该 Broker 接受 `submit_order` 和 `cancel_order`，然后连接撮合引擎，进行撮合，再通过event_bus来讲下单之后的相应返回给RQAlpha。

    AbstractBroker 的存在，意味着实现自定义Broker成为了可能，RQAlpha 可以通过实现任何交易接口对接RQAlpha的Broker来对其进行支持。
    """

    def submit_order(self, order):
        """
        【Required】

        扩展的 Broker 需要实现下单函数，RQAlpha 会根据用户调用的下单 API 自定生成一个 :class:`~Order` 的对象作为参数传入。

        注意：订单处理的返回不是通过该函数的返回值，而是需要触发对应的事件， 如下所示:

        ..  code-block:: python3
            :linenos:

            Environment.get_instance().event_bus.publish_event(Events.TRADE, account, trade)

        :type order: :class:`~Order`
        """
        raise NotImplementedError

    def cancel_order(self, order):
        """
        【Required】

        扩展的 Broker 需要实现撤单函数，其他与submit_order类似

        :param order: 订单
        :type order: :class:`~Order`
        """
        raise NotImplementedError

    def before_trading(self):
        """
        【Optional】

        Broker 很多场景下需要接收到 :class:`~AbstractEventSource` 触发的 `before_trading` 事件进行一些交易前处理。

        自定义 Broker 如果有需要的话，可以实现该函数
        """
        raise NotImplementedError

    def after_trading(self):
        """
        【Optional】

        与 before_trading类似，可以通过实现该函数，在交易结束后进行相关的处理。
        """
        raise NotImplementedError

    def get_open_orders(self):
        """
        【Required】

        RQAlpha 需要获取当天的订单数据，

        :return: list[:class:`~Order`]
        """
        raise NotImplementedError

    def get_accounts(self):
        """
        [Required]

        程序启动后，RQAlpha 会从 Broker 查询账户信息。

        Broker 定义为虚拟券商的代理，RQAlpha默认是支持混合策略的(比如包含股票、期货、期权)，因此Broker生成的的账户需要支持混合策略结构，可以参考 RQAlpha 的 DefaultBroker具体账户的定义和实现。

        :return: dict[:class:`Account`]
        """
        raise NotImplementedError

    def update(self, calendar_dt, trading_dt, bar_dict):
        """
        【Optional】

        data source 在数据和时间更新后，RQAlpha 会调用 Broker 的 `update` 函数，来传入对应的时间和数据。

        Broker 根据自己的业务场景来选择是否实现，比如说自带撮合引擎的Broker，会通过 `update` 函数来触发撮合。

        :param calendar_dt: 实际时间
        :param trading_dt: 交易时间
        :param bar_dict: dict[:class:`~BarObject`]
        """
        raise NotImplementedError


class AbstractMod(object):
    def start_up(self, env, mod_config):
        raise NotImplementedError

    def tear_down(self, code, exception=None):
        raise NotImplementedError


class AbstractPersistProvider(object):
    def store(self, key, value):
        raise NotImplementedError

    def load(self, key):
        raise NotImplementedError


class Persistable(with_metaclass(abc.ABCMeta)):
    @abc.abstractmethod
    def get_state(self):
        """
        :return: bytes
        """
        raise NotImplementedError

    @abc.abstractmethod
    def set_state(self, state):
        """
        :param state: bytes
        :return:
        """
        raise NotImplementedError

    @classmethod
    def __subclasshook__(cls, C):
        if cls is Persistable:
            if (any("get_state" in B.__dict__ for B in C.__mro__) and
                    any("set_state" in B.__dict__ for B in C.__mro__)):
                return True
        return NotImplemented





# -*- coding: utf-8 -*-
#
# Copyright 2017 Ricequant, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum


class Events(Enum):
    POST_SYSTEM_INIT = 'post_system_init'
    POST_USER_INIT = 'post_user_init'

    POST_UNIVERSE_CHANGED = 'post_universe_changed'

    PRE_BEFORE_TRADING = 'pre_before_trading'
    BEFORE_TRADING = 'before_trading'
    POST_BEFORE_TRADING = 'post_before_trading'

    PRE_BAR = 'pre_bar'
    BAR = 'bar'
    POST_BAR = 'post_bar'

    PRE_TICK = 'pre_tick'
    TICK = 'tick'
    POST_TICK = 'post_tick'

    PRE_SCHEDULED = 'pre_scheduled'
    POST_SCHEDULED = 'post_scheduled'

    PRE_AFTER_TRADING = 'pre_after_trading'
    AFTER_TRADING = 'after_trading'
    POST_AFTER_TRADING = 'post_after_trading'

    PRE_SETTLEMENT = 'pre_settlement'
    SETTLEMENT = 'settlement'
    POST_SETTLEMENT = 'post_settlement'

    ORDER_NEW = 'order_new'     # 新产生了一个 order
    ORDER_CREATION_PASS = 'order_creation_pass'
    ORDER_CREATION_REJECT = 'order_creation_reject'
    ORDER_CANCELLATION_PASS = 'order_cancellation_pass'
    ORDER_CANCELLATION_REJECT = 'order_cancellation_reject'
    ORDER__UNSOLICITED_UPDATE = 'order_unsolicited_update'

    TRADE = 'trade'

    ON_LINE_PROFILER_RESULT = 'on_line_profiler_result'



# -*- coding: utf-8 -*-
#
# Copyright 2017 Ricequant, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .benchmark_account import BenchmarkAccount
from .stock_account import StockAccount
from .future_account import FutureAccount
from .mixed_account import MixedAccount
from ...const import ACCOUNT_TYPE


def init_accounts(config):
    accounts = {}
    start_date = config.base.start_date
    total_cash = 0
    for account_type in config.base.account_list:
        if account_type == ACCOUNT_TYPE.STOCK:
            stock_starting_cash = config.base.stock_starting_cash
            accounts[ACCOUNT_TYPE.STOCK] = StockAccount(config, stock_starting_cash, start_date)
            total_cash += stock_starting_cash
        elif account_type == ACCOUNT_TYPE.FUTURE:
            future_starting_cash = config.base.future_starting_cash
            accounts[ACCOUNT_TYPE.FUTURE] = FutureAccount(config, future_starting_cash, start_date)
            total_cash += future_starting_cash
        else:
            raise NotImplementedError
    if config.base.benchmark is not None:
        accounts[ACCOUNT_TYPE.BENCHMARK] = BenchmarkAccount(config, total_cash, start_date)

    return accounts
